# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2024/8/15 11:12
import asyncio
import os
import selectors
import sys
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Optional

import _posixshmem


base_path = Path(__file__).parent
sys.path.append(base_path.parent.as_posix())
from gaterpc.utils import (
    UnixEPollEventLoopPolicy, msg_pack, msg_unpack,
    run_in_executor, to_bytes,
)


flag = os.O_CREAT | os.O_EXCL | os.O_RDWR


def read(sel: selectors.DefaultSelector, fd, end_tag):
    sel.register(fd, selectors.EVENT_READ)
    last_raw = b""
    cont = 1
    try:
        while cont:
            events = sel.select(3)
            # print(f"events: {events}")
            # for key, mask in events:
            try:
                raw = os.read(fd, 100)
                # print(raw)
                if raw == last_raw:
                    print(f"raw is last_raw")
                    cont = 0
                elif raw.endswith(end_tag):
                    cont = 0
                else:
                    last_raw = raw
                # log = msg_unpack(raw)
                # print(log)
                # if log["gtid"] == 100:
                #     break
            except Exception as e:
                print(e)
    finally:
        sel.unregister(fd)


async def aread(fd, end_tag):
    sel = selectors.DefaultSelector()
    loop = asyncio.get_running_loop()

    waitt: Optional[asyncio.Future] = None
    def _read():
        nonlocal waitt
        waitt = loop.create_future()
        d = os.read(fd, 100)
        waitt.set_result(d)

    # loop.add_reader(fd, _read)
    sel.register(fd, selectors.EVENT_READ)
    try:
        last_raw = b""
        cont = 1
        while cont:
            # if waitt:
            #     print(f"read data")
            #     raw = await asyncio.wait_for(waitt, 1)
            #     print(raw)
            #     if raw == last_raw:
            #         print(f"raw is last_raw")
            #         cont = 0
            #     elif raw.endswith(end_tag):
            #         print(f"stream end")
            #         cont = 0
            #     else:
            #         last_raw = raw
            # await asyncio.sleep(0)
            events = sel.select(1)
            for key, mask in events:
                if mask == selectors.EVENT_READ:
                    raw = os.read(fd, 100)
                    print(raw)
                    if raw == last_raw:
                        print("raw is last_raw")
                        cont = 0
                    elif raw.endswith(end_tag):
                        print("stream end")
                        cont = 0
                    else:
                        last_raw = raw
    except Exception as e:
        print(e)
    finally:
        sel.unregister(fd)
        # loop.remove_reader(fd)


def aread_call(fd, end_tag):
    try:
        raw = os.read(fd, 100)
        print(raw)
        if raw.endswith(end_tag):
            print(f"read end.")
            return
    except Exception as e:
        print(e)


def write(fd, end_tag):
    sel = selectors.DefaultSelector()
    sel.register(fd, selectors.EVENT_WRITE)
    i = 0
    while i <= 100:
        events = sel.select(1)
        for key, mask in events:
            if key.fileobj == fd:
                log = {
                    "gtid": (i := i + 1),
                    "action": "update",
                    "key": "tttxxxhhh",
                    "value": {
                        "name": "hostname",
                        "ip": "1.1.1.1",
                        "stat": "running" if i % 2 else "stop",
                        "remote_hosts": ["hostname1", "hostname2", "hostname3"]
                    }
                }
                os.write(fd, msg_pack(log))
    os.write(fd, end_tag)
    print("write end.")


def test():
    r, w = os.pipe()
    print(r, type(r))
    print(w, type(w))
    end_tag = to_bytes("HugeDataEnd")
    sel = selectors.DefaultSelector()
    with ProcessPoolExecutor() as pool:
        fu = pool.submit(write, w, end_tag)
        read(sel, r, end_tag)
    os.close(w)
    os.close(r)


async def atest():
    print(f"event loop policy: {asyncio.get_event_loop_policy()}")
    loop = asyncio.get_running_loop()
    print(f"loop type: {type(loop)}")
    r, w = os.pipe()
    end_tag = to_bytes("HugeDataEnd")
    executor = ProcessPoolExecutor()
    fu = run_in_executor(loop, executor, write, w, end_tag)
    # loop.add_reader(r, aread_call, r, end_tag)
    await aread(r, end_tag)
    # await asyncio.sleep(1)
    await fu
    os.close(r)
    os.close(w)
    # await t
    # if t.exception():
    #     raise t.exception()


if __name__ == '__main__':
    # test()
    asyncio.set_event_loop_policy(UnixEPollEventLoopPolicy())
    asyncio.run(atest())



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2024/1/25 11:29
import asyncio

from gaterpc.global_settings import Settings
from gaterpc.utils import StreamReply


async def testp(sr):
    for i in range(10):
        await sr.asend(i)
    await sr.asend(Settings.STREAM_END_MESSAGE)


async def testg(sr):
    async for i in sr:
        print(f"get value: {i}")


async def test():
    loop = asyncio.get_event_loop()
    sr = StreamReply(maxsize=0, timeout=2)
    gt = loop.create_task(testg(sr))
    pt = loop.create_task(testp(sr))
    await gt
    await pt


def main():
    asyncio.run(test())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2023/5/22 16:36
import asyncio
import atexit
import bz2
import copy
import functools
import inspect
import io
import lzma
import os
import platform
import queue
import selectors
import struct
import sys
import threading
import time
import warnings

import zlib

from logging import Handler, getLogger
from multiprocessing import Manager
from multiprocessing.managers import (
    ArrayProxy, DictProxy, ListProxy,
    ValueProxy,
)
from multiprocessing.shared_memory import SharedMemory
from multiprocessing.resource_tracker import unregister
from importlib import import_module
from pathlib import Path
from functools import _make_key, partial, update_wrapper

from collections import deque, namedtuple
from collections.abc import AsyncGenerator, Callable, Generator
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from typing import Any, Optional, TypeVar, Union, ByteString, overload

import msgpack

from .exceptions import BadGzip, HugeDataException, RemoteException


logger = getLogger("commands")
SyncManager = Manager()


class UnixEPollEventLoopPolicy(asyncio.unix_events.DefaultEventLoopPolicy):
    _loop_factory = asyncio.unix_events.SelectorEventLoop

    def new_event_loop(self):
        return self._loop_factory(selectors.EpollSelector())


def singleton(cls):
    instances = dict()

    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances.get(cls)

    return get_instance


def resolve_module_attr(s: str):
    name = s.split('.')
    module_name = name.pop(0)
    try:
        found = import_module(module_name)
        for frag in name:
            try:
                found = getattr(found, frag)
            except AttributeError:
                import_module(f"{module_name}.{frag}")
                found = getattr(found, frag)
        return found
    except ImportError as e:
        v = ValueError(f"Cannot resolve {s}: {e}")
        raise v from e


def check_socket_addr(socket_addr: Optional[str]) -> Optional[str]:
    if socket_addr is None:
        return socket_addr
    system = platform.system().lower()
    proto = socket_addr.split("://")[0]
    if proto == "ipc" and system == "windows":
        raise SystemError("ipc protocol does not work on Windows.")
    if proto not in ("inproc", "ipc", "tcp", "pgm", "epgm"):
        raise SystemError(f"The protocol \"{proto}\" is not supported.")
    return socket_addr


def to_bytes(s: Union[str, bytes, float, int]) -> bytes:
    if isinstance(s, bytes):
        return s
    if isinstance(s, str):
        return s.encode("utf-8")
    fmt = "!d"
    if isinstance(s, int):
        if -128 <= s <= 127:
            fmt = "!b"
        elif -32768 <= s <= 32767:
            fmt = "!h"
        elif -2147483648 <= s <= 2147483647:
            fmt = "!i"
    return struct.pack(fmt, s)


def from_bytes(b: bytes) -> Union[str, float, int]:
    fmt = {
        1: "!b",
        2: "!h",
        4: "!i",
        8: "!d"
    }
    if b_len := len(b) in fmt:
        try:
            return struct.unpack(fmt[b_len], b)[0]
        except struct.error:
            pass
    return b.decode("utf-8")


def encrypt(data):
    # 在进程池执行器中执行
    hash_str = data
    return hash_str


def decrypt(hash_str):
    # 在进程池执行器中执行
    data = hash_str
    return data


async def generator_to_agenerator(generator: Generator) -> AsyncGenerator:
    for element in generator:
        yield element


async def throw_exception_agenerator(ag: AsyncGenerator, exc: BaseException):
    try:
        await ag.athrow(exc)
    except StopAsyncIteration:
        pass


class Empty:

    def __bool__(self):
        return False


empty = Empty()

_CacheInfo = namedtuple("CacheInfo", ["hits", "misses", "maxsize", "currsize"])


class LRUCache:
    PREV, NEXT, KEY, RESULT = 0, 1, 2, 3

    def __init__(self, maxsize: int):
        self.maxsize = maxsize
        self.hits = self.misses = 0
        self.full = False
        self.root = []
        self.root[:] = [self.root, self.root, None, None]
        self.cache = dict()

    def __contains__(self, item):
        if item in self.cache:
            return True
        return False

    def __setitem__(self, key, value):
        if key in self.cache:
            return
        if self.full:
            oldroot = self.root
            oldroot[self.KEY] = key
            oldroot[self.RESULT] = value
            self.root = oldroot[self.NEXT]
            oldkey = self.root[self.KEY]
            oldresult = self.root[self.RESULT]
            self.root[self.KEY] = self.root[self.RESULT] = None
            del self.cache[oldkey]
            self.cache[key] = oldroot
        else:
            last = self.root[self.PREV]
            link = [last, self.root, key, value]
            last[self.NEXT] = self.root[self.PREV] = self.cache[key] = link
            self.full = len(self.cache) >= self.maxsize

    def __getitem__(self, item):
        if (link := self.cache.get(item, None)) is not None:
            link_prev, link_next, _key, result = link
            link_prev[self.NEXT] = link_next
            link_next[self.PREV] = link_prev
            last = self.root[self.PREV]
            last[self.NEXT] = self.root[self.PREV] = link
            link[self.PREV] = last
            link[self.NEXT] = self.root
            self.hits += 1
            return result
        self.misses += 1
        raise KeyError

    def __repr__(self):
        return (f"hits: {self.hits}, misses: {self.misses},"
                f" maxsize: {self.maxsize}, cache size: {len(self.cache)}")


def _coroutine_lru_cache_wrapper(user_function, maxsize, typed, _CacheInfo):
    loop = asyncio.get_event_loop()
    sentinel = object()
    make_key = _make_key
    PREV, NEXT, KEY, RESULT = 0, 1, 2, 3

    cache = {}
    hits = misses = 0
    full = False
    cache_get = cache.get
    cache_len = cache.__len__
    lock = threading.RLock()
    root = []
    root[:] = [root, root, None, None]

    if maxsize == 0:

        async def wrapper(*args, **kwds):
            nonlocal misses
            misses += 1
            result = await user_function(*args, **kwds)
            return result

    elif maxsize is None:

        async def wrapper(*args, **kwds):
            nonlocal hits, misses
            key = make_key(args, kwds, typed)
            result = cache_get(key, sentinel)
            if result is not sentinel:
                hits += 1
                return await result
            cache[key] = f = loop.create_future()
            misses += 1
            result = await user_function(*args, **kwds)
            f.set_result(result)
            return result

    else:

        async def wrapper(*args, **kwds):
            nonlocal root, hits, misses, full
            key = make_key(args, kwds, typed)
            with lock:
                link = cache_get(key)
                if link is not None:
                    link_prev, link_next, _key, result = link
                    link_prev[NEXT] = link_next
                    link_next[PREV] = link_prev
                    last = root[PREV]
                    last[NEXT] = root[PREV] = link
                    link[PREV] = last
                    link[NEXT] = root
                    hits += 1
                    return result
                misses += 1
            f = loop.create_future()
            with lock:
                if key in cache:
                    pass
                elif full:
                    oldroot = root
                    oldroot[KEY] = key
                    oldroot[RESULT] = f
                    root = oldroot[NEXT]
                    oldkey = root[KEY]
                    oldresult = root[RESULT]
                    root[KEY] = root[RESULT] = None
                    del cache[oldkey]
                    cache[key] = oldroot
                else:
                    last = root[PREV]
                    link = [last, root, key, f]
                    last[NEXT] = root[PREV] = cache[key] = link
                    full = (cache_len() >= maxsize)

            f.set_result(await user_function(*args, **kwds))
            return await f

    def cache_info():
        """Report cache statistics"""
        with lock:
            return _CacheInfo(hits, misses, maxsize, cache_len())

    def cache_clear():
        """Clear the cache and cache statistics"""
        nonlocal hits, misses, full
        with lock:
            cache.clear()
            root[:] = [root, root, None, None]
            hits = misses = 0
            full = False

    wrapper.cache_info = cache_info
    wrapper.cache_clear = cache_clear
    return wrapper


def coroutine_lru_cache(maxsize=128, typed=False):
    if isinstance(maxsize, int):
        # Negative maxsize is treated as 0
        if maxsize < 0:
            maxsize = 0
    elif asyncio.iscoroutinefunction(maxsize) and isinstance(typed, bool):
        # The user_function was passed in directly via the maxsize argument
        user_function, maxsize = maxsize, 128
        wrapper = _coroutine_lru_cache_wrapper(
            user_function, maxsize, typed, _CacheInfo
        )
        wrapper.cache_parameters = lambda: {'maxsize': maxsize, 'typed': typed}
        return update_wrapper(wrapper, user_function)
    elif maxsize is not None:
        raise TypeError(
            'Expected first argument to be an integer, a callable, or None'
        )

    def decorating_function(user_function):
        wrapper = _coroutine_lru_cache_wrapper(
            user_function, maxsize, typed, _CacheInfo
        )
        wrapper.cache_parameters = lambda: {'maxsize': maxsize, 'typed': typed}
        return update_wrapper(wrapper, user_function)

    return decorating_function


class Validator:

    def __init__(self, error_message=None):
        if error_message:
            self.error_message = error_message

    def __call__(self, value):
        return value


class TypeValidator(Validator):
    error_message = "require {require_type}."

    def __init__(self, require_type, error_message=None):
        self.require_type = require_type
        super().__init__(error_message)

    def __call__(self, value):
        if not isinstance(value, self.require_type):
            raise TypeError(
                self.error_message.format(require_type=self.require_type)
            )
        return value


def ensure_mkdir(p: Union[str, Path]):
    if isinstance(p, str):
        p = Path(p)
    p.mkdir(parents=True, exist_ok=True)
    return p


_I = TypeVar("_I")
_PV = TypeVar("_PV")


class LazyAttribute:

    def __init__(
        self, raw: Any = empty,
        render: Callable[[_I, _PV], Any] = None,
        process: Callable[[_I, _PV], Any] = None
    ):
        self._raw = raw
        self._process = process
        self._render = render

    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner=None):
        if self._render:
            return self._render(instance, self._raw)
        return self._raw

    def __set__(self, instance, value):
        if self._process:
            self._raw = self._process(instance, value)
        else:
            self._raw = value


class QueueListener(object):
    _sentinel = None

    def __init__(
        self, handlers: Union[list[Handler], tuple[Handler]] = None,
        respect_handler_level=True
    ):
        self.respect_handler_level = respect_handler_level
        self.queue = queue.Queue()
        self.handlers = dict()
        self._thread = None
        if handlers:
            self.add_handlers(handlers)
        atexit.register(self.stop)
        self.start()

    def add_handlers(self, handlers: Union[list[Handler], tuple[Handler]]):
        for handler in handlers:
            self.handlers[handler.get_name()] = handler

    def remove_handler(self, handler: Handler):
        self.handlers.pop(handler.get_name())

    def dequeue(self):
        return self.queue.get()

    def prepare(self, record):
        return record

    def handle(self, record):
        record = self.prepare(record)
        handler = self.handlers.get(record.h_name)
        if handler:
            if not self.respect_handler_level:
                process = True
            else:
                process = record.levelno >= handler.level
            if process:
                handler.handle(record)

    def enqueue_sentinel(self):
        self.queue.put_nowait(self._sentinel)

    def _monitor(self):
        q = self.queue
        has_task_done = hasattr(q, 'task_done')
        while True:
            try:
                record = self.dequeue()
                if record is self._sentinel:
                    if has_task_done:
                        q.task_done()
                    break
                self.handle(record)
                if has_task_done:
                    q.task_done()
            except queue.Empty:
                break
            except Exception as e:
                warnings.warn(e.__repr__())

    def start(self):
        self._thread = t = threading.Thread(target=self._monitor)
        t.daemon = True
        t.start()

    def stop(self):
        if self._thread:
            self.enqueue_sentinel()
            self._thread.join()
            self._thread = None


class AQueueListener(object):
    _sentinel = None

    def __init__(
        self, handlers: Union[list[Handler], tuple[Handler]] = None,
        respect_handler_level=True
    ):
        self.respect_handler_level = respect_handler_level
        self.queue = asyncio.Queue()
        self.handlers = dict()
        self._task: Optional[asyncio.Task] = None
        if handlers:
            self.add_handlers(handlers)
        atexit.register(self.stop)
        # self._loop.call_soon(self.start)
        self.start()

    def add_handlers(self, handlers: Union[list[Handler], tuple[Handler]]):
        for handler in handlers:
            self.handlers[handler.get_name()] = handler

    def remove_handler(self, handler: Handler):
        self.handlers.pop(handler.get_name())

    async def dequeue(self):
        return await self.queue.get()

    def prepare(self, record):
        return record

    async def handle(self, record):
        record = self.prepare(record)
        handler = self.handlers.get(record.h_name)
        if handler:
            if not self.respect_handler_level:
                process = True
            else:
                process = record.levelno >= handler.level
            if process:
                if asyncio.iscoroutinefunction(handler.handle):
                    await handler.handle(record)
                else:
                    handler.handle(record)

    def enqueue_sentinel(self):
        self.queue.put_nowait(self._sentinel)

    async def _monitor(self):
        q = self.queue
        has_task_done = hasattr(q, 'task_done')
        while True:
            try:
                record = await self.dequeue()
                if record is self._sentinel:
                    if has_task_done:
                        q.task_done()
                    break
                await self.handle(record)
                if has_task_done:
                    q.task_done()
            except queue.Empty:
                break
            except Exception as e:
                warnings.warn(e.__repr__())

    def start(self):
        self._task = asyncio.create_task(self._monitor())

    def stop(self):
        if self._task:
            self.enqueue_sentinel()
            # self._task.cancel()


class AQueueHandler(Handler):

    def __init__(
        self, *args,
        handler_class: Union[Handler, str] = None,
        listener: str = "gaterpc.utils.singleton_aqueue_listener",
        **kwargs
    ):
        Handler.__init__(self)
        if isinstance(handler_class, str):
            handler_class = resolve_module_attr(handler_class)
        self.handler = handler_class(*args, **kwargs)
        self._name = f"QueueHandler{id(self.handler)}"
        self.listener = resolve_module_attr(listener)()
        self.h_name = f"{self._name}-handler"
        self.handler.set_name(self.h_name)
        self.listener.add_handlers((self.handler,))

    def prepare(self, record):
        msg = self.format(record)
        record = copy.copy(record)
        record.message = msg
        record.msg = msg
        record.args = None
        record.exc_info = None
        record.exc_text = None
        record.stack_info = None
        record.h_name = self.h_name
        return record

    def enqueue(self, record):
        self.listener.queue.put_nowait(record)

    def emit(self, record):
        try:
            self.enqueue(self.prepare(record))
        except Exception as e:
            self.handleError(record)
            warnings.warn(e.__repr__())

    def close(self):
        if self.listener:
            self.listener.remove_handler(self.handler)
        super().close()


singleton_aqueue_listener = singleton(AQueueListener)
singleton_queue_listener = singleton(QueueListener)


class MessagePack(object):
    """
    用于配置全局的msgpack
    """

    def __init__(self):
        self.prepare_pack: Optional[Callable] = None
        self.unpack_object_hook: Optional[Callable] = None
        self.unpack_object_pairs_hook: Optional[Callable] = None
        self.unpack_list_hook: Optional[Callable] = None

    def __setattr__(self, key, value):
        if key in (
                "prepare_pack",
                "unpack_object_hook",
                "unpack_object_pairs_hook",
                "unpack_list_hook"
        ) and value and not isinstance(value, Callable):
            raise TypeError(f"`{key}` is not callable")
        super().__setattr__(key, value)

    def __delattr__(self, item):
        if item in (
                "prepare_pack",
                "unpack_object_hook",
                "unpack_object_pairs_hook",
                "unpack_list_hook"
        ):
            setattr(self, item, None)
        super().__delattr__(item)

    def loads(self, data: bytes) -> Any:
        obj = msgpack.unpackb(
            data,
            object_hook=self.unpack_object_hook,
            object_pairs_hook=self.unpack_object_pairs_hook,
            list_hook=self.unpack_list_hook
        )
        return obj

    def dumps(self, obj) -> ByteString:
        data = msgpack.packb(obj, default=self.prepare_pack)
        return data


class MsgPackError(ValueError):
    pass


class MsgUnpackError(ValueError):
    pass


message_pack = MessagePack()


def msg_pack(obj: Any) -> ByteString:
    if isinstance(obj, ValueProxy):
        obj = obj.get()
    elif isinstance(obj, ArrayProxy):
        obj = obj.tolist()
    elif isinstance(obj, (DictProxy, ListProxy)):
        obj = obj._getvalue()
    try:
        data = message_pack.dumps(obj)
    except Exception as e:
        raise MsgPackError(f"{obj} pack failed.").with_traceback(
            e.__traceback__
        )
    return data


def msg_unpack(data: ByteString) -> Any:
    try:
        obj = message_pack.loads(data)
    except Exception as e:
        raise MsgUnpackError(f"{data} unpack failed.").with_traceback(
            e.__traceback__
        )
    return obj


def ensure_shm_unlink(shm: SharedMemory):
    try:
        if shm.buf is not None:
            shm.buf.release()
        shm.close()
        shm.unlink()
    except FileNotFoundError:
        try:
            unregister(shm._name, "shared_memory")
        except KeyError:
            pass
    except (KeyError, ImportError):
        pass


class StreamReply(AsyncGenerator):

    def __init__(self, end_message, maxsize=0, timeout=0):
        try:
            self._loop = asyncio.get_running_loop()
        except RuntimeError:
            self._loop = asyncio.get_event_loop()
        self.end_message = end_message
        self.replies = asyncio.Queue(maxsize, loop=self._loop)
        self.timeout = timeout
        self._exit = False
        self._pause = False

    def __aiter__(self):
        return self

    def is_exit(self):
        return self._exit

    async def __anext__(self):
        if self._exit:
            raise GeneratorExit
        if self.timeout:
            value = await asyncio.wait_for(
                self.replies.get(), self.timeout, loop=self._loop
            )
        else:
            value = self.replies.get_nowait()
        self.replies.task_done()
        if value == self.end_message:
            await asyncio.sleep(0.1)
            self._exit = True
            raise StopAsyncIteration
        if self._pause and isinstance(value, BaseException):
            self._pause = False
            raise value
        return value

    async def asend(self, value):
        if value is not None:
            await self.replies.put(value)
        return

    async def athrow(self, __type, value=None, traceback=None):
        self._pause = True
        if traceback:
            await self.replies.put(__type.with_traceback(traceback))
        elif isinstance(value, BaseException):
            await self.replies.put(value)
        else:
            if value:
                __type = __type(value)
            await self.replies.put(__type)

    async def aclose(self):
        if not self._exit:
            self._exit = True


class GzipCompressor(object):

    def __init__(self, compresslevel):
        self.compresslevel = compresslevel
        self.crc = zlib.crc32(b"")
        self.size = 0
        self._add_header = True
        self._compressor = zlib.compressobj(
            self.compresslevel,
            zlib.DEFLATED,
            -zlib.MAX_WBITS,
            zlib.DEF_MEM_LEVEL,
            zlib.Z_DEFAULT_STRATEGY
        )

    def _generate_header(self, mtime=None) -> bytes:
        header = b"\037\213\010\000"
        if not mtime:
            mtime = time.time()
        header += struct.pack("<L", int(mtime))
        if self.compresslevel == 9:
            header += b"\002"
        elif self.compresslevel == 1:
            header += b"\004"
        else:
            header += b"\000"
        header += b"\377"
        return header

    def compress(self, data):
        if isinstance(data, bytes):
            length = len(data)
        else:
            data = memoryview(data)
            length = data.nbytes
        if length > 0:
            self.crc = zlib.crc32(data, self.crc)
            data = self._compressor.compress(data)
            self.size += length
            if self._add_header:
                data = self._generate_header() + data
                self._add_header = False
        return data

    def flush(self):
        data = self._compressor.flush()
        data += struct.pack("<L", self.crc)
        data += struct.pack("<L", self.size & 0xffffffff)
        return data


class GzipDecompressor(object):

    def __init__(self):
        self._buffer = b""
        self._crc = zlib.crc32(b"")
        self._stream_size = 0
        self._new_member = True
        self._decompressor = zlib.decompressobj(-zlib.MAX_WBITS)

    def _remove_header(self, data):
        index = 0
        magic = data[index: (index := index + 2)]
        if magic != b"\037\213":
            raise BadGzip("Not a gzipped data.")
        method, flag, self.last_mtime = struct.unpack(
            "<BBIxx", data[index: (index := index + 8)]
        )
        if method != 8:
            raise BadGzip("Unknown conmpression method")
        if flag & 4:
            extra_len, = struct.unpack(
                "<H", data[index: (index := index + 2)]
            )
            index += extra_len
        if flag & 8:
            while True:
                s = data[(index := index + 1)]
                if not s or s == b'\000':
                    break
        if flag & 16:
            while True:
                s = data[(index := index + 1)]
                if not s or s == b'\000':
                    break
        if flag & 2:
            index += 2
        return data[index:]

    @property
    def eof(self):
        return self._decompressor.eof

    @property
    def needs_input(self):
        return self._decompressor.unconsumed_tail == b""

    def _read_eof(self):
        if buffer := self._buffer[:8]:
            crc32, isize = struct.unpack("<II", buffer)
            if crc32 != self._crc:
                raise BadGzip(f"CRC check failed {hex(crc32)} != {hex(self._crc)}")
            elif isize != (self._stream_size & 0xffffffff):
                raise BadGzip("Incorrect length of data produced")

    def decompress(self, data, max_length=0):
        if max_length < 0:
            max_length = 0
        if self._new_member:
            data = self._remove_header(data)
            self._new_member = False
        data += self._buffer
        data = self._decompressor.decompress(data, max_length)
        self._buffer = self._decompressor.unconsumed_tail
        if data:
            self._crc = zlib.crc32(data, self._crc)
        self._stream_size += len(data)
        return data


def _incremental_compress(
    compress_module: str, compress_level: int,
    in_data: Union[str, int], out_data: int, end_tag: bytes,
    blksize: int = 1000
):
    """
    :param compress_module: 压缩模块
    :param compress_level: 压缩级别
    :param in_data: 输入数据，管道读取端的文件描述符，或者 SharedMemory 的名
    :param out_data: 输出数据，管道写入端的文件描述符
    :param end_tag: 结束标记
    :param blksize: 每一块数据的大小
    """
    sel = selectors.DefaultSelector()
    sel.register(out_data, selectors.EVENT_WRITE)
    write_key = (sel.get_key(out_data), selectors.EVENT_WRITE)
    if isinstance(in_data, str):
        shm = SharedMemory(name=in_data)
        shm_index = 0

        def read_data(events):
            nonlocal shm_index
            data = shm.buf[
                   shm_index:
                   (shm_index := shm_index + io.DEFAULT_BUFFER_SIZE)
                   ].tobytes()
            if data:
                return data
            else:
                return end_tag

        def clean():
            try:
                shm.close()
            except BufferError as e:
                # logger.warning(f"_incremental_compress.clean: {e}")
                pass

    elif isinstance(in_data, int):
        sel.register(in_data, selectors.EVENT_READ)
        read_key = (sel.get_key(in_data), selectors.EVENT_READ)

        def read_data(events):
            if read_key in events:
                dl = os.read(in_data, 2)
                d = os.read(in_data, int.from_bytes(dl, sys.byteorder))
                return d
            return None

        def clean():
            sel.unregister(in_data)

    else:
        raise TypeError("in_data")

    try:
        if compress_module == "gzip":
            compressor = GzipCompressor(compress_level)
        elif compress_module == "bz2":
            compressor = bz2.BZ2Compressor(compress_level)
        elif compress_module == "lzma":
            compressor = lzma.LZMACompressor(preset=compress_level)
        else:
            compress_module = resolve_module_attr(compress_module)
            compressor = compress_module.compressor(compress_level)
        buffer = b""
        stream_end = False
        finish = False
        while 1:
            events = sel.select(1)
            if stream_end and not finish:
                buffer += compressor.flush()
                finish = True
            elif (len(buffer) > blksize) or (stream_end and buffer):
                if write_key in events:
                    c_d, buffer = (
                        buffer[:blksize],
                        buffer[blksize:]
                    )
                    c_d = len(c_d).to_bytes(2, sys.byteorder) + c_d
                    os.write(out_data, c_d)
            elif not stream_end:
                raw_d = read_data(events)
                if raw_d == end_tag:
                    stream_end = True
                elif raw_d:
                    buffer += compressor.compress(raw_d)
            else:
                break
    except Exception as error:
        raise HugeDataException(error)
    finally:
        os.write(out_data, len(end_tag).to_bytes(2, sys.byteorder) + end_tag)
        clean()
        sel.unregister(out_data)
        sel.close()


def _incremental_decompress(
    compress_module: str,
    in_data: Union[str, int], out_data: int,
    end_tag: bytes,
    max_length=-1, blksize=1000
):
    """
    :param compress_module: 压缩模块
    :param in_data: 输入数据，管道读取端的文件描述符，或者 SharedMemory 的名
    :param out_data: 输出数据，管道写入端的文件描述符
    :param end_tag: 结束标记
    :param max_length: 解压缩参数
    :param blksize: 每一块数据的大小
    """
    sel = selectors.DefaultSelector()
    sel.register(out_data, selectors.EVENT_WRITE)
    write_key = (sel.get_key(out_data), selectors.EVENT_WRITE)
    if isinstance(in_data, str):
        shm = SharedMemory(name=in_data)
        shm_index = 0

        def read_data(events):
            nonlocal shm_index
            data = shm.buf[
                   shm_index:
                   (shm_index := shm_index + io.DEFAULT_BUFFER_SIZE)
                   ].tobytes()
            if data:
                return data
            else:
                return end_tag

        def clean():
            try:
                shm.close()
            except BufferError as e:
                # logger.warning(f"_incremental_decompress.clean: {e}")
                pass

    elif isinstance(in_data, int):
        sel.register(in_data, selectors.EVENT_READ)
        read_key = (sel.get_key(in_data), selectors.EVENT_READ)

        def read_data(events):
            if read_key in events:
                dl = os.read(in_data, 2)
                return os.read(in_data, int.from_bytes(dl, sys.byteorder))
            return None

        def clean():
            sel.unregister(in_data)

    else:
        raise TypeError("in_data")

    try:
        if compress_module == "gzip":
            decompressor = GzipDecompressor()
        elif compress_module == "bz2":
            decompressor = bz2.BZ2Decompressor()
        elif compress_module == "lzma":
            decompressor = lzma.LZMADecompressor()
        else:
            compress_module = resolve_module_attr(compress_module)
            decompressor = compress_module.decompressor()
        buffer = b""
        stream_end = False
        while 1:
            events = sel.select(1)
            if write_key in events:
                if buffer:
                    d_d, buffer = (
                        buffer[:blksize],
                        buffer[blksize:]
                    )
                    d_d = len(d_d).to_bytes(2, sys.byteorder) + d_d
                    os.write(out_data, d_d)
            if not buffer:
                if decompressor.eof:
                    break
                if decompressor.needs_input:
                    if stream_end:
                        raise OSError("The compressed data is incomplete")
                    c_d = read_data(events)
                else:
                    c_d = b""
                if c_d and c_d.endswith(end_tag):
                    c_d = c_d[:-1 * len(end_tag)]
                    stream_end = True
                if c_d is not None:
                    buffer += decompressor.decompress(c_d, max_length)
    except Exception as error:
        raise HugeDataException(error)
    finally:
        os.write(out_data, len(end_tag).to_bytes(2, sys.byteorder) + end_tag)
        clean()
        sel.unregister(out_data)
        sel.close()


class HugeData:

    def __init__(
        self,
        end_tag,
        except_tag,
        *,
        data: Optional[bytes] = None,
        compress_module: str = "gzip",
        compress_level: int = 1,
        blksize: int = 1000,
        timeout: float = 1,
    ):
        self.sel = selectors.DefaultSelector()
        self.end_tag = end_tag
        self.except_tag = except_tag
        self.data: Union[SharedMemory, queue.Queue]
        if data:
            self.data = SharedMemory(create=True, size=len(data))
            self.data.buf[:len(data)] = data
        else:
            self.data: tuple = os.pipe()
        self.compress_module: str = compress_module
        self.compress_level: int = compress_level
        self.blksize: int = blksize
        self._p: tuple = os.pipe()
        self._write_buffer = b""
        self._write_end = False
        self._remote_exception = None
        self.timeout = timeout

        atexit.register(self.destroy)

    def destroy(self):
        if isinstance(self.data, SharedMemory):
            ensure_shm_unlink(self.data)
        elif isinstance(self.data, tuple):
            os.close(self.data[0])
            os.close(self.data[1])
            self.data = None
        if self._p:
            os.close(self._p[0])
            os.close(self._p[1])
            self._p = None
        if self.sel:
            self.sel.close()
            self.sel = None
        atexit.unregister(self.destroy)

    def __del__(self):
        self.destroy()

    def add_data(self, data: bytes):
        if not isinstance(self.data, tuple):
            raise TypeError("data")
        if self._write_end:
            raise RuntimeError("pipe close")
        self._write_buffer += data

    def flush(self):
        self._write_end = True

    def _write_data1(self):
        if self._write_buffer:
            r_d, self._write_buffer = (
                self._write_buffer[:self.blksize],
                self._write_buffer[self.blksize:]
            )
            r_d = len(r_d).to_bytes(2, sys.byteorder) + r_d
            os.write(self.data[1], r_d)
        elif self._write_end:
            os.write(
                self.data[1],
                len(self.end_tag).to_bytes(2, sys.byteorder) + self.end_tag
            )

    async def compress(
        self, loop: asyncio.AbstractEventLoop
    ) -> AsyncGenerator[Union[bytes, tuple]]:
        clean_writer = False
        if isinstance(self.data, SharedMemory):
            args = (
                self.compress_module, self.compress_level,
                self.data.name, self._p[1], self.end_tag
            )
        else:
            args = (
                self.compress_module, self.compress_level,
                self.data[0], self._p[1], self.end_tag
            )

            loop.add_writer(self.data[1], self._write_data1)
            clean_writer = True

        func = partial(
            _incremental_compress, *args,
            blksize=self.blksize
        )
        read_q = asyncio.Queue()

        def _read():
            dl = os.read(self._p[0], 2)
            d = os.read(self._p[0], int.from_bytes(dl, sys.byteorder))
            read_q.put_nowait(d)

        loop.add_reader(self._p[0], _read)
        process_executor = ProcessPoolExecutor()
        f = loop.run_in_executor(process_executor, func)
        try:
            await asyncio.sleep(0)
            cont = 1
            read_num = 0
            while cont:
                await asyncio.sleep(0)
                if self._remote_exception:
                    remote_exception = msg_unpack(self._remote_exception)
                    raise RemoteException(remote_exception)
                try:
                    data = await asyncio.wait_for(read_q.get(), self.timeout)
                    read_q.task_done()
                except asyncio.TimeoutError:
                    data = b""
                if data:
                    read_num += 1
                    if data.endswith(self.end_tag):
                        yield data[:-1 * len(self.end_tag)]
                        cont = 0
                    else:
                        yield data
        finally:
            loop.remove_reader(self._p[0])
            if clean_writer:
                loop.remove_writer(self.data[1])
            if f.done():
                exc = f.exception()
            else:
                exc = None
                f.cancel()
            process_executor.shutdown(False, cancel_futures=True)
            if exc:
                raise exc

    async def decompress(
        self, loop: asyncio.AbstractEventLoop, max_length=-1
    ) -> AsyncGenerator[Union[bytes, tuple]]:
        clean_writer = False
        if isinstance(self.data, SharedMemory):
            args = (
                self.compress_module,
                self.data.name, self._p[1], self.end_tag
            )
        else:
            args = (
                self.compress_module,
                self.data[0], self._p[1], self.end_tag
            )

            loop.add_writer(self.data[1], self._write_data1)
            clean_writer = True

        func = partial(
            _incremental_decompress, *args,
            max_length=max_length, blksize=self.blksize
        )
        read_q = asyncio.Queue()

        def _read():
            dl = os.read(self._p[0], 2)
            d = os.read(self._p[0], int.from_bytes(dl, sys.byteorder))
            read_q.put_nowait(d)

        loop.add_reader(self._p[0], _read)
        process_executor = ProcessPoolExecutor()
        f = loop.run_in_executor(process_executor, func)
        try:
            await asyncio.sleep(0)
            cont = 1
            read_num = 0
            while cont:
                if self._remote_exception:
                    remote_exception = msg_unpack(self._remote_exception)
                    raise RemoteException(remote_exception)
                try:
                    data = await asyncio.wait_for(read_q.get(), self.timeout)
                    read_q.task_done()
                except asyncio.TimeoutError:
                    data = b""
                if data:
                    read_num += 1
                    if data.endswith(self.end_tag):
                        yield data[:-1 * len(self.end_tag)]
                        cont = 0
                    else:
                        yield data
                await asyncio.sleep(0)
        finally:
            loop.remove_reader(self._p[0])
            if clean_writer:
                loop.remove_writer(self.data[1])
            if f.done():
                exc = f.exception()
            else:
                exc = None
                f.cancel()
            process_executor.shutdown(False, cancel_futures=True)
            if exc:
                raise exc


@overload
def interface(method_or_executor: Callable):
    """
    :param method_or_executor: 给暴露出去的方法添加属性
    """


@overload
def interface(method_or_executor: str):
    """
    :param method_or_executor: 当方法为同步方法时，用于执行的执行器偏好，
      可选 thread 或 process 或 none，none 表示不使用执行器
    """


def interface(method_or_executor):
    if inspect.isfunction(method_or_executor):
        method_or_executor.__interface__ = True
        method_or_executor.__executor__ = "none"
        return method_or_executor
    elif (
            isinstance(method_or_executor, str)
            and method_or_executor in ("thread", "process", "none")
    ):
        if inspect.iscoroutinefunction(method_or_executor):
            raise RuntimeError

        def wrapper(methode):
            methode.__interface__ = True
            methode.__executor__ = method_or_executor
            return methode

        return wrapper
    raise TypeError(method_or_executor.__class__)


def run_in_executor(loop, executor, func, *args, **kwargs) -> asyncio.Future:
    func_call = functools.partial(func, *args, **kwargs)
    return loop.run_in_executor(executor, func_call)



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2024/2/28 16:10
import asyncio
import random
import time
from collections import deque
from functools import lru_cache
from uuid import uuid4

from gaterpc.utils import coroutine_lru_cache


A = uuid4().hex.encode("utf-8")
B = uuid4().hex.encode("utf-8")
C = uuid4().hex.encode("utf-8")
D = [
        [b"NULL", ],
        [b"PLAIN", b"testa", b"testap"],
        [b"PLAIN", b"testb", b"testbp"],
        [b"PLAIN", b"testc", b"testcp"],
        [b"PLAIN", b"testu", b"testp"],
    ]


class CacheCoroutine(object):
    def __init__(self, maxsize=128):
        self.maxsize = maxsize

    def __call__(self, coroutine):
        @lru_cache(maxsize=self.maxsize)
        def wrapper(*args, **kwargs) -> asyncio.Task:
            loop = asyncio.get_event_loop()
            t = loop.create_task(coroutine(*args, *kwargs))
            return t
        return wrapper


@coroutine_lru_cache
async def test_cache(a: bytes, b: bytes, c: bytes, *d: list) -> bool:
    await asyncio.sleep(2)
    hits = 0
    if a == A:
        print("hit A")
        hits += 1
    if b == B:
        print("hit B")
        hits += 1
    if c == C:
        print("hit C")
        hits += 1
    try:
        if d[0] == b"PLAIN" and d[1] == b"testu" and d[2] == b"testp":
            if hits == 3:
                print("hit D")
            return True
        else:
            return False
    except Exception:
        return False


async def atest():
    loop = asyncio.get_event_loop()
    ti = 0
    aws = deque()
    for i in range(3000):
        if not ti:
            a, b, c = A, B, C
            d = D[-1]
            ti = random.randint(1, 10)
        else:
            a = uuid4().hex.encode("utf-8")
            b = uuid4().hex.encode("utf-8")
            c = uuid4().hex.encode("utf-8")
            d = D[random.randint(0, 4)]
        ti -= 1
        aws.append(t := loop.create_task(test_cache(a, b, c, *d)))
        # t.add_done_callback(aws.remove)
    await asyncio.gather(*aws)
    print(test_cache.cache_info())


def main():
    asyncio.run(atest())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2023/6/8 11:04
import asyncio
import threading
from asyncio import events


_global_lock = threading.Lock()


class _LoopBoundMixin:
    _loop: asyncio.AbstractEventLoop = None

    def _get_loop(self) -> asyncio.AbstractEventLoop:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = events.get_event_loop()

        if self._loop is None:
            with _global_lock:
                if self._loop is None:
                    self._loop = loop
        elif loop is not self._loop:
            raise RuntimeError(f'{self!r} is bound to a different event loop')
        return loop



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2024/2/23 15:55
import asyncio
import hashlib
import inspect
import sys
from logging import getLogger
from pathlib import Path
from traceback import format_exception

import zmq.constants as z_const
import zmq.auth

base_path = Path(__file__).parent
sys.path.append(base_path.parent.as_posix())

from gaterpc.core import Client
from gaterpc.global_settings import Settings
from gaterpc.utils import check_socket_addr

import testSettings


curve_dir = Path(__file__).parent.joinpath("curvekey/")
if curve_dir.exists():
    g_public, _ = zmq.auth.load_certificate(
        curve_dir.joinpath("gate.key")
    )
    cw_public, cw_secret = zmq.auth.load_certificate(
        curve_dir.joinpath("cw.key_secret")
    )
else:
    g_public = b""
    cw_public = cw_secret = b""

Settings.configure("USER_SETTINGS", testSettings)
logger = getLogger("commands")


test_t_set = set()


def cleanup_t(t: asyncio.Task):
    test_t_set.remove(t)
    try:
        if t.exception():
            logger.debug(f"{t.get_name()}.exception: {t.exception()}")
    except asyncio.CancelledError:
        pass


async def client(frontend_addr):
    loop = asyncio.get_running_loop()
    if cw_secret:
        Settings.ZMQ_SOCK.update(
            {
                z_const.CURVE_SECRETKEY: cw_secret,
                z_const.CURVE_PUBLICKEY: cw_public,
                z_const.CURVE_SERVERKEY: g_public,
            }
        )
    Settings.MESSAGE_MAX = 10000  # 参考ZMQ HWM
    gr_cli = Client()
    await gr_cli.connect(check_socket_addr(frontend_addr))
    await asyncio.sleep(5)
    logger.info(await gr_cli.GateRPC.get_interfaces())
    logger.info(gr_cli._remote_services)
    logger.info(await gr_cli.Gate.query_service("GateRPC"))
    try:
        i = 10000
        st = loop.time()
        while i or test_t_set:
            if i:
                co = gr_cli.test("a", "b", "c", time=gr_cli._loop.time())
                test_t_set.add(
                    t := loop.create_task(co)
                )
                t.add_done_callback(cleanup_t)
                # logger.info("=====================================================")
                # logger.info(f"result of test: {result}")
                # logger.info("=====================================================")
                i -= 1
                if not i:
                    logger.info(f"creat task use time: {loop.time() - st}")
            await asyncio.sleep(0)
        logger.info(f"func test use time: {loop.time() - st}")

        i = 10000
        st = loop.time()
        while i or test_t_set:
            if i:
                co = gr_cli.atest("d", "e", "f", time=gr_cli._loop.time())
                test_t_set.add(
                    t := loop.create_task(co)
                )
                t.add_done_callback(cleanup_t)
                # logger.info("=====================================================")
                # logger.info(f"result of atest: {result}")
                # logger.info("=====================================================")
                i -= 1
                if not i:
                    logger.info(f"creat task use time: {loop.time() - st}")
            await asyncio.sleep(0)
        logger.info(f"func atest use time: {loop.time() - st}")

        rw_i = 0
        st = loop.time()
        async for i in await gr_cli.test_agenerator(10):
            # logger.info("=====================================================")
            # logger.info(f"get i for agen: {i}")
            # logger.info("=====================================================")
            if i != rw_i:
                raise RuntimeError
            rw_i += 1
            await asyncio.sleep(0)
        print(f"rw_i: {rw_i}, use time: {loop.time() - st}")
        st = loop.time()
        dd = await gr_cli.test_huge_data()
        logger.info(f"test_huge_data use time: {loop.time() - st}")
        logger.info("=====================================================")
        logger.info(f"length of dd: {len(dd)}")
        logger.info(
            f"dd sha256: {hashlib.sha256(dd.encode('utf-8')).hexdigest()}"
        )
        logger.info("=====================================================")
        # logger.info("check heartbeat")
        # await gr_cli._recv_task
    except Exception as e:
        logger.info("*****************************************************")
        for line in format_exception(*sys.exc_info()):
            logger.error(line)
        logger.info("*****************************************************")
    finally:
        logger.info(
            f"the length of client's replies: {len(gr_cli.replies)}"
        )
        logger.info(
            f"the length of client's replies: {len(gr_cli.replies)}"
        )
        gr_cli.close()


async def test(frontend_addr):
    Settings.DEBUG = 1
    Settings.setup()
    await client(frontend_addr)


if __name__ == '__main__':
    try:
        frontend_addr = sys.argv[1]
        asyncio.run(test(frontend_addr))
    except IndexError:
        print("require frontend_addr.")



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2024/3/1 9:21
import asyncio
import sys
import time

import zmq.constants as z_const
import zmq.asyncio as z_aio
from gaterpc.core import Context


async def test_req(name: str, route_addr: str, ctx=None):
    if not ctx:
        ctx = Context()
    socket = ctx.socket(z_const.REQ, z_aio.Socket)
    socket.set(z_const.IDENTITY, name.encode("utf-8"))
    socket.set_hwm(3000)
    socket.set(z_const.CONNECT_TIMEOUT, 3000)
    socket.set(z_const.SNDTIMEO, 3000)
    socket.set(z_const.RCVTIMEO, 3000)
    print("connect timeout", socket.get(z_const.CONNECT_TIMEOUT))
    print("recv timeout", socket.get(z_const.RCVTIMEO))
    print("send timeout", socket.get(z_const.SNDTIMEO))
    print("heartbeat timeout", socket.get(z_const.HEARTBEAT_TIMEOUT))
    try:
        print(time.time())
        socket.connect(route_addr)
        print(time.time())
        for i in range(10):
            req = [b"test req", str(i).encode("utf-8")]
            print(time.time())
            await socket.send_multipart(req)
            print(time.time())
            replies = await socket.recv_multipart()
            print(time.time())
            print(replies)
    finally:
        print(time.time())


if __name__ == '__main__':
    argv = sys.argv[1:]
    name = argv[0]
    route_addr = argv[1]
    asyncio.run(test_req(name, route_addr))



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2024/2/5 15:45
import asyncio
# import uvloop
import time
import concurrent.futures


async def test():
    from gaterpc.core import Client
    from gaterpc.global_settings import Settings

    # Settings.EVENT_LOOP_POLICY = uvloop.EventLoopPolicy()
    Settings.DEBUG = 1
    Settings.setup()
    loop = asyncio.get_event_loop()
    loop.slow_callback_duration = 0.01
    gr_cli = Client(
        broker_addr="tcp://127.0.0.1:777",
        # zap_mechanism=Settings.ZAP_MECHANISM_PLAIN.decode("utf-8"),
        # zap_credentials=(
        #     Settings.ZAP_PLAIN_DEFAULT_USER,
        #     Settings.ZAP_PLAIN_DEFAULT_PASSWORD
        # )
    )
    # await asyncio.sleep(2)
    start = time.time()
    ts = [
        loop.create_task(gr_cli.Gate.query_service("GateRPC"))
        for i in range(2000)
    ]
    for t in ts:
        await t
    end = time.time()
    gr_cli.close()
    return end - start


def test_thread():
    use_time = asyncio.run(test())
    if use_time > 1:
        print(use_time)


def main():
    # with concurrent.futures.ProcessPoolExecutor() as executor:
    #     fs = []
    #     start = time.time()
    #     for i in range(1000):
    #         print(f"{i}", end="\r")
    #         f = executor.submit(test_thread)
    #         fs.append(f)
    #     concurrent.futures.wait(fs, 60)
    #     end = time.time()
    #     print(f"Total time: {end - start}")
    test_thread()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2024/1/8 15:52
import asyncio
import os
import pickle
import random
import secrets
import selectors
import sys
import time
from concurrent.futures.process import ProcessPoolExecutor
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from logging import exception
from pathlib import Path
from traceback import format_tb
from typing import Optional



base_path = Path(__file__).parent
sys.path.append(base_path.parent.as_posix())

from gaterpc.global_settings import Settings
from gaterpc.utils import (
    HugeData, UnixEPollEventLoopPolicy, _incremental_compress,
    _incremental_decompress, msg_pack, run_in_executor,
)
import testSettings


Settings.configure("USER_SETTINGS", testSettings)
Settings.setup()
tc_file = base_path.joinpath("test_compress")
com_module = "bz2"


def test_incremental_compress(hdata: HugeData):
    cd = b""
    os.set_inheritable(hdata._p[1], True)
    args = (
        hdata.compress_module, hdata.compress_level,
        hdata.data.name, hdata._p[1], hdata.end_tag
    )
    for arg in args:
        print(arg)
        pickle.dumps(arg)
    executor = ProcessPoolExecutor()
    func = partial(_incremental_compress, *args)
    fu = executor.submit(func)
    # _incremental_compress(*args)
    sel = selectors.DefaultSelector()
    sel.register(hdata._p[0], selectors.EVENT_READ)
    check_fu = True
    _continue = 1
    try:
        while _continue:
            if check_fu:
                if fu.done():
                    check_fu = False
                    executor.shutdown()
                    if exc := fu.exception():
                        raise exc
            events = sel.select(1)
            for key, mask in events:
                if mask == selectors.EVENT_READ:
                    d = os.read(hdata._p[0], hdata.blksize)
                    if d.endswith(hdata.end_tag):
                        cd += d[:-1 * len(hdata.end_tag)]
                        _continue = 0
                    else:
                        cd += d
        return cd
    finally:
        sel.unregister(hdata._p[0])
        with tc_file.open("wb") as f:
            f.write(cd)
        hdata.destroy()


def test_incremental_decompress():
    dd = b""
    with tc_file.open("rb") as f:
        ts = f.read()
    hdata = HugeData(
        Settings.HUGE_DATA_END_TAG,
        Settings.HUGE_DATA_EXCEPT_TAG,
        compress_module=com_module
    )
    os.set_inheritable(hdata.data[0], True)
    os.set_inheritable(hdata._p[1], True)
    args = (
        hdata.compress_module,
        hdata.data[0], hdata._p[1],
        hdata.end_tag
    )
    for arg in args:
        pickle.dumps(arg)
    executor = ProcessPoolExecutor()
    func = partial(
        _incremental_decompress, *args, max_length=1000
    )
    fu = executor.submit(func)
    ts_len = len(ts)
    print(f"ts len: {ts_len}")
    sel = selectors.DefaultSelector()
    sel.register(hdata._p[0], selectors.EVENT_READ)
    sel.register(hdata.data[1], selectors.EVENT_WRITE)
    index = 0
    _write = True
    _stream_end = False
    check_fu = True
    _continue = 1
    try:
        while _continue:
            if check_fu:
                if fu.done():
                    check_fu = False
                    executor.shutdown()
                    if exc := fu.exception():
                        raise exc
            events = sel.select(1)
            for key, mask in events:
                if mask == selectors.EVENT_WRITE and _write:
                    wdata = ts[index: (index := index+1000)]
                    if wdata:
                        os.write(hdata.data[1], wdata)
                    else:
                        os.write(hdata.data[1], hdata.end_tag)
                        _write = False
                if mask == selectors.EVENT_READ:
                    d = os.read(hdata._p[0], hdata.blksize)
                    if d.endswith(hdata.end_tag):
                        d = d[:-1 * len(hdata.end_tag)]
                        dd += d
                        _continue = 0
                    else:
                        dd += d
        return dd
    finally:
        sel.unregister(hdata._p[0])
        sel.unregister(hdata.data[1])
        print("hd.destroy")
        hdata.destroy()


async def compress_agenerator(hdata:HugeData):
    sel = selectors.DefaultSelector()
    loop = asyncio.get_running_loop()
    args = (
        hdata.compress_module, hdata.compress_level,
        hdata.data.name, hdata._p[1], hdata.end_tag
    )
    func = partial(_incremental_compress, *args)
    _wait: Optional[asyncio.Future] = None
    _wait_id = ""

    def _read():
        nonlocal _wait, _wait_id
        _wait = loop.create_future()
        _wait_id = id(_wait)
        _wait.set_result(os.read(hdata._p[0], hdata.blksize))

    sel.register(hdata._p[0], selectors.EVENT_READ)
    read_key = (sel.get_key(hdata._p[0]), selectors.EVENT_READ)
    executor = ProcessPoolExecutor()
    try:
        f = loop.run_in_executor(executor, func)
        check_fu = True
        cont = 1
        while cont:
            if check_fu:
                if f.done():
                    print("future run end")
                    executor.shutdown()
                    check_fu = False
                    if exc := f.exception():
                        raise exc
            # if _wait:
            #     if _wait_id != id(_wait):
            #         raise RuntimeError("id error")
            #     try:
            #         data = await asyncio.wait_for(_wait, hdata.get_timeout)
            #         if data.endswith(hdata.end_tag):
            #             yield data[:-1 * len(hdata.end_tag)]
            #             break
            #         else:
            #             yield data
            #     except asyncio.TimeoutError:
            #         pass
            #     _wait = None
            # await asyncio.sleep(0)
            events = sel.select(1)
            if read_key in events:
                data = os.read(hdata._p[0], hdata.blksize)
                if data.endswith(hdata.end_tag):
                    yield data[:-1 * len(hdata.end_tag)]
                    cont = 0
                else:
                    yield data
    # except Exception as error:
    #     except_info = (
    #             str(error.__class__),
    #             str(error),
    #             "".join(format_tb(error.__traceback__))
    #         )
    #     print(except_info)
    #     await asyncio.sleep(1)
    finally:
        # loop.remove_reader(hdata._p[0])
        sel.unregister(hdata._p[0])


def send_rawd(hdata: HugeData, raw_d):
    hdata.add_data(raw_d)
    hdata.flush()
    print(f"send end.")


async def incremental_compress(hdata: HugeData, raw_d=b""):
    loop = asyncio.get_running_loop()
    cd = b""
    texecutor = ThreadPoolExecutor()
    send_t = None
    if raw_d:
        send_t = loop.run_in_executor(texecutor, send_rawd, hdata, raw_d)
    try:
        async for dd in hdata.compress(loop):
            cd += dd
        if send_t:
            await send_t
        with tc_file.open("wb") as f:
            f.write(cd)
        cdl = len(cd)
        return cdl
    finally:
        texecutor.shutdown(False, cancel_futures=True)


async def decompress_agenerator(hdata: HugeData):
    sel = selectors.DefaultSelector()
    loop = asyncio.get_running_loop()
    with tc_file.open("rb") as f:
        ts = f.read()
    args = (
        hdata.compress_module,
        hdata.data[0], hdata._p[1],
        hdata.end_tag
    )
    func = partial(_incremental_decompress, *args, max_length=1000)
    r_wait: Optional[asyncio.Future] = None
    r_wait_id = ""
    w_wait: Optional[asyncio.Future] = None
    w_wait_id = ""

    def _write():
        nonlocal w_wait, w_wait_id
        w_wait = loop.create_future()
        w_wait_id = id(w_wait)
        w_wait.set_result(True)

    def _read():
        nonlocal r_wait, r_wait_id
        r_wait = loop.create_future()
        r_wait_id = id(r_wait)
        r_wait.set_result(os.read(hdata._p[0], hdata.blksize))

    sel.register(hdata._p[0], selectors.EVENT_READ)
    read_key = (sel.get_key(hdata._p[0]), selectors.EVENT_READ)
    sel.register(hdata.data[1], selectors.EVENT_WRITE)
    write_key = (sel.get_key(hdata.data[1]), selectors.EVENT_WRITE)
    executor = ProcessPoolExecutor()
    index = 0
    ts_len = len(ts)
    _write = True
    _stream_end = False
    print(f"ts len: {ts_len}")
    try:
        f = loop.run_in_executor(executor, func)
        check_fu = True
        cont = 1
        while cont:
            data = b""
            if check_fu:
                if f.done():
                    check_fu = False
                    executor.shutdown()
                    if exc := f.exception():
                        raise exc
            events = sel.select(1)
            # print(events)
            if read_key in events:
                data = os.read(hdata._p[0], hdata.blksize)
                if data.endswith(hdata.end_tag):
                    data = data[:-1 * len(hdata.end_tag)]
                    cont = 0
            if write_key in events and _write:
                wdata = ts[index: (index := index+1000)]
                if wdata:
                    os.write(hdata.data[1], wdata)
                else:
                    os.write(hdata.data[1], hdata.end_tag)
                    _write = False
            if data:
                yield data
    finally:
        sel.unregister(hdata.data[1])
        sel.unregister(hdata._p[0])


def send_dd(hdata: HugeData, dd):
        # events = sel.select(1)
        # if write_key in events:
    hdata.add_data(dd)
    hdata.flush()
    print("send end.")


async def incremental_decompress(hdata: HugeData, dd=b""):
    loop = asyncio.get_running_loop()
    raw_d = b""
    # executor = ProcessPoolExecutor()
    texecutor = ThreadPoolExecutor()
    # t = loop.create_task(send_dd(hd))
    send_t = None
    if dd:
        send_t = loop.run_in_executor(texecutor, send_dd, hdata, dd)
    st = loop.time()
    try:
        async for _d in hdata.decompress(loop, max_length=1024):
            raw_d += _d
        if send_t:
            await send_t
    # await asyncio.gather(dd, tt)
    # async for data in hd.decompress(1000):
    #     if throw:
    #         raise RuntimeError(data)
    #     if data == hd.except_tag:
    #         throw = True
    #     else:
    #         dd += data
        return raw_d
    finally:
        # executor.shutdown(False, cancel_futures=True)
        texecutor.shutdown(False, cancel_futures=True)
        print(f"raw_d len: {len(raw_d)}")
    # except Exception as error:
    #     except_info = (
    #         str(error.__class__),
    #         str(error),
    #         "".join(format_tb(error.__traceback__))
    #     )
    #     print(except_info)


def main(mode="all"):
    # asyncio.set_event_loop_policy(UnixEPollEventLoopPolicy())
    bs = ""
    if mode in ("all", "compress"):
        i = random.randint(10000, 30000)
        while i:
            bs += secrets.token_hex()
            i -= 1
        print(f"raw bs len: {len(bs)}")
        bs = msg_pack(bs)
        bsl = len(bs)
        print(f"bs len: {bsl}")
        print(asyncio.get_event_loop_policy())
        hd = HugeData(
            Settings.HUGE_DATA_END_TAG,
            Settings.HUGE_DATA_EXCEPT_TAG,
            # data=bs,
            compress_module=com_module,
            compress_level=Settings.HUGE_DATA_COMPRESS_LEVEL
        )
        print("asyncio.run(incremental_compress(hd)")
        cdl = asyncio.run(incremental_compress(hd, bs), debug=True)
        hd.destroy()
        print(f"cd len: {cdl}")
        print(cdl / bsl)

    if mode in ("all", "decompress"):
        with tc_file.open("rb") as f:
            ts = f.read()
        hd = HugeData(
            Settings.HUGE_DATA_END_TAG,
            Settings.HUGE_DATA_EXCEPT_TAG,
            data=ts,
            compress_module=com_module,
        )
        print(hd.data)
        print("asyncio.run(incremental_decompress(hd)")
        dd = asyncio.run(incremental_decompress(hd,), debug=True)
        hd.destroy()
        print(f"dd len: {len(dd)}")
        if bs:
            is_ = dd == bs
            print(f"dd == bs: {is_}")
            if not is_:
                raise RuntimeError

    print("main end")


if __name__ == '__main__':
    if len(sys.argv) > 1:
        main(sys.argv[1])
    else:
        main()
    # test_incremental_compress(hd)
    # test_incremental_decompress()
    # snapshot = tracemalloc.take_snapshot()
    # top_stats = snapshot.statistics("lineno")
    # for stat in top_stats:
    #     print(stat)



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2024/8/23 13:51
import asyncio
import hashlib
import secrets
import sys
from logging import getLogger
from pathlib import Path

import zmq.auth
import zmq.constants as z_const


base_path = Path(__file__).parent
sys.path.append(base_path.parent.as_posix())

from gaterpc.global_settings import Settings
from gaterpc.core import (
    AsyncZAPService, Context, Worker, Service, AMajordomo,
)
from gaterpc.utils import (
    HugeData, UnixEPollEventLoopPolicy, interface, msg_pack, run_in_executor,
    to_bytes,
)
import testSettings


curve_dir = Path(__file__).parent.joinpath("curvekey/")
if curve_dir.exists():
    g_public, _ = zmq.auth.load_certificate(
        curve_dir.joinpath("gate.key")
    )
    cw_public, cw_secret = zmq.auth.load_certificate(
        curve_dir.joinpath("cw.key_secret")
    )
else:
    g_public = b""
    cw_public = cw_secret = b""

Settings.configure("USER_SETTINGS", testSettings)
Settings.DEBUG = 1
logger = getLogger("commands")
s = ""
i = 10000
while i:
    s += secrets.token_hex()
    i -= 1

s_256 = hashlib.sha256(s.encode("utf-8")).hexdigest()


class GRWorker(Worker):
    @interface
    async def aconcurrency(self, *args, **kwargs):
        return b""

    @interface
    async def atest(self, *args, **kwargs):
        return {
            "name": "async atest",
            "args": args,
            "kwargs": kwargs,
            "loop_time": self._loop.time()
        }

    @interface("none")
    def test(self, *args, **kwargs):
        return {
            "name": "test",
            "args": args,
            "kwargs": kwargs,
            "loop_time": self._loop.time()
        }

    @interface("thread")
    def test_generator(self, maximum: int):
        i = 0
        while i < maximum:
            yield i
            i += 1

    @interface
    async def test_agenerator(self, maximum: int):
        i = 0
        while i < maximum:
            await asyncio.sleep(0.1)
            yield i
            i += 1

    @interface
    async def test_huge_data(self):
        # logger.info(f"s sha256: {s_256}")
        try:
            data = await run_in_executor(
                self._loop, self.process_executor, 
                msg_pack, s
            )
            logger.info(f"data len: {len(data)}, sha256: {s_256}")
            hd = HugeData(
            Settings.HUGE_DATA_END_TAG,
            Settings.HUGE_DATA_EXCEPT_TAG,
            data=data
        )
            return hd
        except Exception as e:
            logger.error(e)

    @interface
    async def emit(self, log):
        try:
            lid = log["gtid"]
            return lid
        except Exception as e:
            logger.error(e)
            raise


async def worker(backend_addr=None):
    Settings.setup()
    if cw_secret:
        Settings.ZMQ_SOCK.update({
            z_const.CURVE_SECRETKEY: cw_secret,
            z_const.CURVE_PUBLICKEY: cw_public,
            z_const.CURVE_SERVERKEY: g_public,
        })
    # loop = asyncio.get_event_loop()
    # loop.slow_callback_duration = 0.01
    if backend_addr:
        Settings.WORKER_ADDR = backend_addr
    ctx = Context()
    gr = Service()
    gr_worker = gr.create_worker(
        GRWorker,
        context=ctx,
        # zap_mechanism=Settings.ZAP_MECHANISM_PLAIN,
        # zap_credentials=(
        #     Settings.ZAP_PLAIN_DEFAULT_USER,
        #     Settings.ZAP_PLAIN_DEFAULT_PASSWORD
        # ),
        max_allowed_request=100000
    )
    logger.info(gr_worker.service)
    if gr_worker.service is not gr:
        return False
    logger.info(gr_worker.interfaces)
    try:
        gr_worker._reconnect = True
        gr_worker.run()
        while 1:
            if gr_worker._recv_task:
                if gr_worker._recv_task.done():
                    break
                await gr_worker._recv_task
    finally:
        logger.info(
            f"the length of worker's requests: {len(gr_worker.requests)}"
        )
        await asyncio.sleep(1)
        logger.info(
            f"the length of worker's requests: {len(gr_worker.requests)}"
        )
        gr_worker.stop()


def test(backend_addr=None):
    asyncio.set_event_loop_policy(UnixEPollEventLoopPolicy())
    asyncio.run(worker(backend_addr))


if __name__ == '__main__':
    if len(argv := sys.argv) > 1:
        test(argv[0])
    else:
        test()



import asyncio

from gaterpc.utils import BoundedDict


async def test():
    d = BoundedDict(1000, 3.0)
    await asyncio.gather(
        test_get(d),
        test_pop(d),
        test_set(d),
    )
    d["aset"] = "test sync set"
    await asyncio.sleep(1)
    print(f"get 'aset': {d.get('aset')}")
    print(f"length: {len(d)}")


async def test_get(d):
    aget1000 = await d.aget(1000, "failed")
    print(f"first time aget '1000': {aget1000}")
    await asyncio.sleep(1)
    aget1000 = await d.aget(1000, "failed")
    print(f"second time aget '1000': {aget1000}")
    await asyncio.sleep(1)
    aset = await d.aget("aset", "failed")
    print(f"first time aget 'aset': {aset}")
    sec_aset = await d.aget("aset", "failed")
    print(f"second time aget 'aset': {sec_aset}")


async def test_pop(d):
    apop1000 = await d.apop(1000, "failed")
    print(f"first time apop '1000': {apop1000}")
    apop1000 = await d.apop(1000, "failed")
    print(f"second time apop '1000': {apop1000}")
    await asyncio.sleep(1)
    print(f"popitem: {d.popitem()}")
    aset = await d.apop("aset", "failed")
    print(f"first time apop 'aset' : {aset}")
    sec_aset = await d.apop("aset", "failed")
    print(f"second time apop 'aset' : {sec_aset}")


async def test_set(d):
    for i in range(1000):
        print(f"aset {i}")
        await d.aset(i, i)
        await asyncio.sleep(0)
    _as = asyncio.create_task(d.aset(1000, "test 1000"))
    _as1 = asyncio.create_task(d.aset(1001, "test 1001"))
    print(f"pop: {d.pop(1)}")
    print(f"pop: {d.pop(2)}")
    await asyncio.sleep(3)
    await d.aset("aset", "first time aset.")
    await asyncio.sleep(1)
    await d.aset("aset", "second time aset.")
    await _as
    await _as1


def main():
    asyncio.run(test())
    print(f"test is callable: {callable(test)}")


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2023/8/10 17:38
class ServiceUnAvailableError(Exception):
    def __init__(self, name):
        super().__init__(
            "The {name} Service unavailable.".format(name=name),
        )


class GateUnAvailableError(ServiceUnAvailableError):
    def __init__(self):
        super().__init__("Gate")


class BusyWorkerError(Exception):
    def __init__(self):
        super().__init__("No idle worker.")


class BusyGateError(Exception):
    def __init__(self):
        super().__init__("No idle gate.")


class RemoteException(Exception):
    def __init__(self, except_info):
        if isinstance(except_info, str):
            message = except_info
        elif isinstance(except_info, (tuple, list)):
            except_class, except_value, except_traceback = except_info
            message = (
                "\n"
                "  RemoteTraceback ({except_class}):\n"
                "  {except_value}\n"
                "  {except_traceback}"
            ).format(
                except_class=except_class,
                except_value=except_value,
                except_traceback=except_traceback
            )
        else:
            raise TypeError("except_info must be str or tuple.")
        super().__init__(message)


class BadGzip(OSError):
    """"""


class HugeDataException(OSError):
    """"""


class DictFull(Exception):
    def __init__(self, maxsize):
        message = (
            "The number of keys in the dict reached the maximum of {maxsize}."
        ).format(maxsize=maxsize)
        super().__init__(message)



# -*- coding: utf-8 -*-
# Author      : ShiFan
# Created Date: 2023/12/29 9:12
import asyncio
# import uvloop
import sys
from logging import getLogger
from pathlib import Path
from traceback import format_exception
from typing import Optional, Union
from uuid import uuid4

import zmq.auth
import zmq.constants as z_const


base_path = Path(__file__).parent
sys.path.append(base_path.parent.as_posix())

from gaterpc.global_settings import Settings
from gaterpc.core import (
    AsyncZAPService, Context, Worker, Service, AMajordomo,
)
from gaterpc.utils import (
    HugeData, UnixEPollEventLoopPolicy, interface, msg_pack, msg_unpack,
    run_in_executor,
    to_bytes,
)
import testSettings


curve_dir = Path(__file__).parent.joinpath("curvekey/")
if curve_dir.exists():
    g_public, g_secret = zmq.auth.load_certificate(
        curve_dir.joinpath("gate.key_secret")
    )
    cw_public, _ = zmq.auth.load_certificate(
        curve_dir.joinpath("cw.key")
    )
else:
    g_public = g_secret = b""
    cw_public = b""

Settings.configure("USER_SETTINGS", testSettings)
logger = getLogger("commands")


class TAMajordomo(AMajordomo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clients.add(self.identity)
        self.rw_queue = asyncio.Queue()
        self.rw_task: Optional[asyncio.Task] = None
        self.rw_replies = dict()

    async def request_worker(self):
        await asyncio.sleep(3)
        try:
            rpc_service = self.services[Settings.SERVICE_DEFAULT_NAME]
        except KeyError:
            logger.error("RPC service don't running.")
            await asyncio.sleep(1)
            self.rw_task = self._loop.create_task(self.request_worker())
            return
        st = 0
        i = 0
        while 1:
            if i > 100000:
                break
            try:
                worker = await rpc_service.acquire_idle_worker()
            except asyncio.TimeoutError:
                continue
            if not st:
                st = self._loop.time()
                logger.debug(f"start time: {st}")
            log = {
                    "gtid": i + 1,
                    "action": "update",
                    "key": "tttxxxhhh",
                    "value": {
                        "name": "hostname",
                        "ip": "1.1.1.1",
                        "stat": "running" if i % 2 else "stop",
                        "remote_hosts": ["hostname1", "hostname2", "hostname3"]
                    }
                }
            request_id = to_bytes(uuid4().hex)
            body = (
                msg_pack("emit"),
                msg_pack((log,))
            )
            option = (self.identity, b"", request_id)

            # 要注意并发太多时，对套接字类型的选择和缓冲区的配置，同时适时的让出io
            # 可以适当提高Settings里的ZMQ_SOCk配置里的 z_const.HWM，
            # sysctl -w net.core.wmem_default=33554432
            # sysctl -w net.core.rmem_default=33554432
            # sysctl -w net.core.wmem_max=67108864
            # sysctl -w net.core.rmem_max=67108864

            # 可以顺序等待发送
            # await self.send_to_backend(
            #     worker.identity, Settings.MDP_COMMAND_REQUEST,
            #     option, body
            # )
            # 也可以创建为发送任务，乱序发送
            self.backend_tasks.add(
                t := self._loop.create_task(
                    self.send_to_backend(
                            worker.identity, Settings.MDP_COMMAND_REQUEST,
                            option, body
                        )
                )
            )
            t.add_done_callback(self.cleanup_backend_task)
            if log["gtid"] == 100000:
                logger.debug(f"sent use time: {self._loop.time() - st}")
            self.rw_replies[request_id] = self._loop.create_future()
            i += 1
            await asyncio.sleep(0)

    rw_replies_num = 0

    async def reply_frontend(
        self, client_id: bytes,
        service_name: Union[str, bytes],
        request_id: Union[str, bytes],
        *body
    ):
        if client_id == self.identity:
            if request_id in self.rw_replies:
                # logger.debug(f"result: {msg_unpack(body[0])}")
                self.rw_replies.pop(request_id).set_result(body)
                if not self.rw_replies_num:
                    logger.debug(f"first replies: {self._loop.time()}")
                self.rw_replies_num += 1
                if not self.rw_replies_num % 10000:
                    logger.debug(
                        f"worker replies number: {self.rw_replies_num}, "
                        f"time: {self._loop.time()}"
                    )
        else:
            await super().reply_frontend(
                client_id, service_name, request_id, *body
            )

    async def test_rw(self):
        await asyncio.sleep(5)
        logger.info("start test rw")
        try:
            st = self._loop.time()
            for i in range(100000):
                log = {
                    "gtid": i + 1,
                    "action": "update",
                    "key": "tttxxxhhh",
                    "value": {
                        "name": "hostname",
                        "ip": "1.1.1.1",
                        "stat": "running" if i % 2 else "stop",
                        "remote_hosts": ["hostname1", "hostname2", "hostname3"]
                    }
                }
                await self.rw_queue.put(log)
                await asyncio.sleep(0)
            et = self._loop.time()
            logger.info(f"put use time: {et - st}")
        except Exception as e:
            logger.info("*****************************************************")
            for line in format_exception(*sys.exc_info()):
                logger.error(line)
            logger.info("*****************************************************")
            raise e

    def run(self):
        if not self._broker_task:
            self._broker_task = self._loop.create_task(self._broker_loop())
        if not self.rw_task:
            self.rw_task = self._loop.create_task(self.request_worker())

    def stop(self):
        if self.rw_task:
            if not self.rw_task.cancelled():
                self.rw_task.cancel()
            self.rw_task = None
        if self._broker_task:
            if not self._broker_task.cancelled():
                self._broker_task.cancel()
            self.close()
            self._broker_task = None


async def zap_server(ctx):
    zap = AsyncZAPService(context=ctx)
    zap.configure_plain(
        Settings.ZAP_DEFAULT_DOMAIN,
        {
            Settings.ZAP_PLAIN_DEFAULT_USER: Settings.ZAP_PLAIN_DEFAULT_PASSWORD
        }
    )
    try:
        zap.start()
        await zap._recv_task
    finally:
        zap.stop()


def start_zap(ctx):
    loop = asyncio.new_event_loop()
    loop.slow_callback_duration = 0.01
    loop.run_until_complete(zap_server(ctx))


async def majordomo(
    frontend=None, backend=None, bind_gate=None, connect_gate=None
):
    Settings.setup()
    # loop = asyncio.get_event_loop()
    # loop.slow_callback_duration = 0.01
    if backend:
        Settings.WORKER_ADDR = backend
    Settings.ZAP_REPLY_TIMEOUT = 10.0
    ctx = Context()
    # zipc = Settings.ZAP_ADDR
    # logger.info(f"zap ipc addr: {zipc}")
    # zap = AsyncZAPService(addr=zipc)
    # zap.configure_plain(
    #     Settings.ZAP_DEFAULT_DOMAIN,
    #     {
    #         Settings.ZAP_PLAIN_DEFAULT_USER: Settings.ZAP_PLAIN_DEFAULT_PASSWORD
    #     }
    # )

    gr_majordomo = TAMajordomo(
        context=ctx,
        # gate_zap_mechanism=Settings.ZAP_MECHANISM_PLAIN,
        # gate_zap_credentials=(
        #     Settings.ZAP_PLAIN_DEFAULT_USER,
        #     Settings.ZAP_PLAIN_DEFAULT_PASSWORD
        # )
    )
    if g_secret:
        gr_majordomo.bind_backend(sock_opt={
                z_const.CURVE_SECRETKEY: g_secret,
                z_const.CURVE_PUBLICKEY: g_public,
                z_const.CURVE_SERVER: True,
            })
    else:
        gr_majordomo.bind_backend()
    if frontend:
        if g_secret:
            gr_majordomo.bind_frontend(
                frontend,
                sock_opt={
                    z_const.CURVE_SECRETKEY: g_secret,
                    z_const.CURVE_PUBLICKEY: g_public,
                    z_const.CURVE_SERVER: True,
                }
            )
        else:
            gr_majordomo.bind_frontend(frontend)
    if bind_gate:
        gr_majordomo.bind_gate(bind_gate)
    await asyncio.sleep(1)
    logger.info("start test")
    try:
        # zap.start()
        # gr_majordomo.connect_zap(zap_addr=zipc)
        gr_majordomo.run()
        if connect_gate:
            try:
                await gr_majordomo.connect_gate(connect_gate)
            except Exception as e:
                logger.error(e)
                raise e
        await asyncio.sleep(5)
        # await gr_majordomo.test_rw()
        await gr_majordomo.rw_task
        await gr_majordomo._broker_task
    finally:
        exc = gr_majordomo.rw_task.exception()
        gr_majordomo.stop()
        logger.info(f"request_zap.cache_info: {gr_majordomo.zap_cache}")
        # zap.stop()
        if exc:
            raise exc


def test(
    frontend="ipc:///tmp/gate-rpc/run/c1",
    backend=None, bind_gate=None, connect_gate=None
):
    print(f"frontend: {frontend}, backend: {backend}")
    print(f"bind_gate: {bind_gate}, connect_gate: {connect_gate}")
    asyncio.set_event_loop_policy(UnixEPollEventLoopPolicy())
    asyncio.run(
        majordomo(frontend, backend, bind_gate, connect_gate)
    )


if __name__ == "__main__":
    argv = sys.argv[1:]
    test(*argv)
